
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <conio.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/seginfo.h>
#include <sys/_dskdrvr.h>
#include <sys/fsys.h>
#include <sys/osinfo.h>

#include <qnx/drvdecl.h>
#include <qnx/drvmount.h>
#include <qnx/sys/sdebug.h>
#include <qnx/sys/timeout.h>
#include <qnx/sys/pos.h>

#include "ps2scsi.h"
#include "extern.h"

static char *hex_str = "0123456789abcdef";

check_pos()
{
int i, t;
static char buf[5];
static struct ps2_dinfo ps2s;

    if (ps2_getsys(PS2_SYS_SCSI,&ps2s) != -1) {
       if (ps2s.setupinfo[0] & 0x01) {  /* is present */
           wstr("Fsys.ps2s: system board PS/2 SCSI adapter installed.\n",1);
           return 0;
       }
    }
    
    for (i=0; i < 8; i++) {
        if ((t=ps2_getbd(i,&ps2s)) == PS2_SCSI_CACHE || t == PS2_SCSI_NOCACHE) {
           break;
        }
    }
    if (i == 8) {
       wstr("Fsys.ps2s: PS/2 SCSI not detected\n",2);
       return -1;
    }
    wstr("Fsys.ps2s: expansion board PS/2 SCSI in slot ",1);
    write(1,hex_str + i,1);
    switch ((ps2s.setupinfo[2]>>1) & 007) {
    case    0:    io_addr = 0x3540;    break;
    case    1:    io_addr = 0x3548;    break;
    case    2:    io_addr = 0x3550;    break;
    case    3:    io_addr = 0x3558;    break;
    case    4:    io_addr = 0x3560;    break;
    case    5:    io_addr = 0x3568;    break;
    case    6:    io_addr = 0x3570;    break;
    case    7:    io_addr = 0x3578;    break;
    }
    wstr("at io_address 0x",1);
    buf[0] = hex_str[(io_addr >> 12) & 0xf];
    buf[1] = hex_str[(io_addr >> 8)  & 0xf];
    buf[2] = hex_str[(io_addr >> 4)  & 0xf];
    buf[3] = hex_str[(io_addr >> 0)  & 0xf];
    buf[4] = 0;
    wstr(buf,1);
    wstr("\n",1);
    return 0;
}
/*
 * @@@@ undocumented options:
 *     -L nlines  == number of lines in the debugging version of the driver.
 *     -u         == use the programmable option select to configure.
 */
        
main(int argc, char **argv) 
{
int     opt; 
int usepos=0;
int nlines=16;
    while ((opt=getopt(argc,argv,"a:i:m:n:N:L:u")) != EOF) {
        switch( opt ) {
        case 'a':        /*    Control address */
            io_addr = strtol(optarg,NULL,16);
            usepos = 0;
            break;
        case 'i':        /*    IRQ level */
            hw_irq = atoi( optarg );
            usepos = 0;
            break;
        case 'm':
            MaxIO = strtol(optarg,NULL,0);
            break;
        case 'u':
            usepos = 1;
            break;
        case 'n':
            if (strlen(optarg) < 1) {
                return EXIT_FAILURE;
            }
            if (strlen(optarg) > _DRIVER_NAME_LEN) {
                wstr("error, name too long\n",2);
                return EXIT_FAILURE;
            }
            strcpy(driver_name,optarg);
            break;
        case    'L':
            nlines = strtol(optarg,0,0);
            break;   
        case    'N':
            if (strlen(optarg) < 1) {
                wstr("unique name too short\n",2);
                return EXIT_FAILURE;
            }
            if (strlen(optarg) >= sizeof(version_stamp.name)) {
                wstr("unique name too long\n",2);
                return EXIT_FAILURE;
            }
            strcpy(version_stamp.name,optarg);
            break;
        
        default:
            wstr("Unknown option ignored\n",2);
        }
    }
#ifndef NDEBUG    
{
int debug_sel;
    qnx_osinfo(0,&_osdata);
    debug_sel = (_osdata.primary_monitor & 001) ? MONO_DISP : COLOUR_DISP;
    set_display(MK_FP(debug_sel,2*80*2),nlines,72);
}

#endif

    if (usepos && check_pos() == -1) {
        wstr("Fsys.ps2s: unable to locate IBM SCSI Adapter\n",2);
        return EXIT_FAILURE;
    }
        
    for (opt=0; opt < 32 &&
        (FsysPid = qnx_name_locate(getnid(),fsys_name,0,0)) == -1; opt++)
        ;
    if (FsysPid == -1) {
        perror("name_locate");
        wstr("unable to locate filesystem\n",2);
    }
    get_cpu_speed();

    if (fsys_mount_driver(FsysPid, MAX_NDRIVES, disk, etable,
                    MaxIO, MAX_PARALLEL,stack+STACK_SIZE-1,
                    &version_stamp,_DRVR_DMA) == -1) {
        perror("mount");
        wstr("unable to attach to filesystem\n",2);
        return EXIT_FAILURE;
    }
#ifdef DEBUG
    while (1)   sleep(60);
#endif
    return EXIT_SUCCESS;
}
