
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <i86.h>
#include <conio.h>
#include <sys/name.h>
#include <sys/irqinfo.h>
#include <sys/kernel.h>
#include <sys/seginfo.h>
#include <sys/osinfo.h>
#include <sys/fsys.h>
#include "ps2scsi.h"
#include "extern.h"



#ifdef DEBUG
volatile long   ior_count;
volatile long   proxy_count;
#endif

/* device handler */
struct hd_func  scsimap[PS2_NDEVS];
#define putfunc(_id,_f,_ptr)   (scsimap[(_id)].f=_f,scsimap[(_id)].ptr=_ptr)

/*
 * the ps2 scsi spec says buffers must be on dword boundaries. Since doing
 * this would cause all structures defined in include files to behave
 * unpredictably, we allocate one array which guarantee's enough space to
 * hold all of the structures, on dword boundaries.
 */


/* scsi control block */
struct _scb    __far *scb_ptr;

/* scsi unit capacity */
struct _capacity __far *cap_ptr;

/* transaction status buffer */
struct _trbuf {
	ushort_t        end_status, retries;
	daddr_t         remainder;
	paddr_t         sc_element;
	ushort_t        dstatlen;
	ushort_t        cmdstat;/* hi word = int status, low word = devstat */
	ushort_t        cmdecode;
	ushort_t        rsvd;
	ushort_t        hitrate;
	paddr_t         last_scb;
};

struct _trbuf  __far *tsb_ptr;

/* physical addresses of above */
paddr_t         scb_phys_addr, cap_phys_addr, tsb_phys_addr;


#define MEMPOOL_SIZE    (sizeof(struct _scb) +\
                         sizeof(struct _capacity)+\
                         sizeof(short)*13 +\
                         3 * sizeof(long))

caddr_t        mempool;
dma_t	dbuf;

#define ALIGN(x)    (((x) + 3) & ~(3))

static
init_vars()
{
	caddr_t		p;

	if (alloc_dma_buf(&dbuf, 0, MEMPOOL_SIZE) == -1)
		return -1;
		
	p = dbuf.bufp;
	
	scb_ptr = (struct _scb __far *) p;
	scb_phys_addr = dbuf.phys_addr;

	p += ALIGN(sizeof(struct _scb));
	cap_ptr = (struct _capacity __far *) p;
	cap_phys_addr = dbuf.phys_addr + (p - dbuf.bufp);

	p += ALIGN(sizeof(struct _capacity));

	tsb_ptr = (struct _trbuf __far *) p;
	tsb_phys_addr = dbuf.phys_addr + (p - dbuf.bufp);

}

init_done(int istat, int bstat, caddr_t ptr)
{
	DriverState = 1;
	return 0;
}

DCL_INIT(ps2s_init, proxy, tabsize, cachetab)
{
	int             rc;

	NotifyPid = proxy;
	if (init_vars() == -1)
		return -1;

	if ((rc = alloc_dma_buf(&iobuffer, FsysPid, MaxIO * _BLOCK_SIZE)) == -1) {
		return -1;
	}
	if ((int_id = qnx_hint_attach(hw_irq,
			    MK_FP(my_cs(), &int_handler), my_ds())) == -1) {
		qnx_segment_free(FP_SEG(iobuffer.bufp));
		return -1;
	}
	/*
	 * reset the adapter.  This has the effect of resetting every device
	 * attached.
	 */
	putfunc(ADAPTER_ID, init_done, 0);
	outb(io_addr + CONTROL, HARDWARE_RESET);
	delay(50);
	outb(io_addr + CONTROL, 0);
	outb(io_addr + CONTROL, INT_ENABLE);
	for (rc = 0; DriverState == 0 && rc < 60; rc++)
		delay(1000);
	if (rc == 60)
		return -1;
	outb(io_addr + CONTROL, INT_ENABLE | DMA_ENABLE);
	/*
	 * The adapter is now setup.
	 */
	return 0;
}

DCL_TERM(ps2s_term, flags)
{
	qnx_segment_free(FP_SEG(iobuffer.bufp));
	qnx_hint_detach(int_id);
	return 0;
}

int
finish_open(int istat, int bstat, caddr_t bptr)
{
	int             unit = istat & 0x0f;
	disk_t         *dp = disk + unit;

	dp->d_type = 0;
	dp->d_retval = _DRVR_UNSPEC;
	switch (INTR_STATUS(istat)) {
	case ADAPTER_RESET:
		break;
	case COMMAND_ERROR:
		disp_str("error{");
		disp_hex(tsb_ptr->end_status);
		disp_hex(tsb_ptr->cmdstat);
		disp_hex(tsb_ptr->cmdecode);
		disp_char('}');
		/*
		 * If you get a device not available, give up Otherwise, keep
		 * trying to get it's cap
		 */
		if ((tsb_ptr->cmdecode & 0xff) == 0x10)
			break;
		putfunc(istat & 0x0f, finish_open, 0);
		if (send_command(START_SCB | (istat & 0x0f), scb_phys_addr) != 0) {
			break;
		}
		return 0;
		break;
	case COMMAND_OK:
	case COMMAND_RETRY:
		swab_dword(&dp->d_num_blks, &cap_ptr->nblocks);
		dp->d_type = _HARD;
		dp->d_heads = 64;
		dp->d_sctr_trk = 32;
		dp->d_sctr_cyl = dp->d_heads * dp->d_sctr_trk;
		dp->d_cyls = dp->d_num_blks / dp->d_sctr_cyl;
		dp->d_retval = _DRVR_OK;
		strcpy(dp->d_prefix, driver_name);
		break;
	case ADAPTER_FAILURE:
	case IMMEDIATE_OK:
		break;
	case COMMAND_INVALID:
	case SOFTWARE_FAULT:
		break;
	default:
		disp_str("open failed, bad int status =");
		disp_hex(istat);
	}
	DriverState = 1;
	return 0;
}

DCL_OPEN(ps2s_open, disknum, flags)
{
	disk_t         *dp;
#ifdef DEBUG
	disp_str("open(");
	disp_hex(disknum);
	disp_char(')');
#endif
	if (disknum >= MAX_NDRIVES) {
		return -1;
	}
	(dp = disk + disknum)->d_busy = 1;	/* The drive is busy while we
						 * open it            */
	dp->d_sctr_base = 1;	/* controller numbers blocks starting at 1    */

	/*
	 * Ask the controller for the size of the disk.
	 */
	DriverState = 0;	/* still 0 -- not real commands yet */
	putfunc(disknum, finish_open, 0);
	cap_cmd();
	if (send_command(START_SCB | disknum, scb_phys_addr) == 0) {
		while (DriverState == 0)
			Yield();
	} else {
#ifdef DEBUG
		disp_str("open failed, bad send_command");
#endif
		putfunc(disknum, 0, 0);
		dp->d_retval = -1;
		dp->d_type = 0;
	}
	dp->d_busy = 0;
	DriverState = 1;	/* now we are for real! */
	if (!(flags & _OPEN_SYNC))
		Trigger(NotifyPid);
	return dp->d_retval;
}

DCL_CLOSE(ps2s_close, disknum)
{
	memset(disk + disknum, 0, sizeof(*disk));
	return 0;
}

int
io_done(int istat, int bstat, caddr_t p)
{
	int             psw;
	int             unit = istat & 0x0f;
	disk_t         *dsk = disk + unit;

	psw = disable();
	outb(SYS_PORT_A, inb(SYS_PORT_A) & ~DISK_LIGHT);
	restore(psw);
#ifdef DEBUG
#if DEBUG > 2
	disp_hex(unit);
#endif
#endif
	if (INTR_STATUS(istat) == 1 || INTR_STATUS(istat) == 5) {
		dsk->d_retval = 0;
		dsk->d_nblks = NumBlocks;
		if (Function == READ_DATA) {
#ifdef COPY_DMA_BUF
			buf_to_cache(CachePtr, iobuffer.bufp, NumBlocks);
#else
			dsk->d_dma_seg = FP_SEG(iobuffer.bufp);
#endif
		}
	} else {
		dsk->d_retval = -1;
	}
#ifdef DEBUG
#if DEBUG > 1
	disp_str("io done(");
	disp_hex(dsk->d_retval);
	disp_char(')');
#else
	if (dsk->d_stat) {
		disp_hex(dsk->d_stat);
	}
	disp_char(')');
#endif
	proxy_count++;
#endif
	dsk->d_busy = 0;
	return NotifyPid;
}


#ifdef COPY_DMA_BUF
cbufp_t         CachePtr;
#endif



DCL_IO(ps2s_io, disknum, mode, nblocks, bp)
{
	disk_t         *dp;
	int             psw;
	if (disknum >= MAX_NDRIVES || nblocks > MaxIO) {
		return -1;
	}
#ifdef DEBUG
#if DEBUG > 1
	disp_str("io(");
	disp_hex(disknum);
	disp_char(')');
#else
	disp_char('(');
#endif
	ior_count++;
#endif
#ifdef COPY_DMA_BUF
	CachePtr = bp;
#endif
	(dp = disk + disknum)->d_busy = 1;
	dp->d_retval = -1;
	dp->d_nblks = 0;
	if (mode & _WRITE_BLK) {
		Function = WRITE_DATA;
		cache_to_buf(iobuffer.bufp, bp, nblocks);
	} else {
		Function = READ_DATA;
	}
	NumBlocks = nblocks;

	io_cmd(Function, bp->b_block - 1, iobuffer.phys_addr, nblocks);
	psw = disable();
	outb(SYS_PORT_A, inb(SYS_PORT_A) | DISK_LIGHT);
	restore(psw);
	putfunc(disknum, io_done, 0);
	if ((send_command(START_SCB | disknum, scb_phys_addr)) == 0)
		return 0;
	return -1;
}


INT_PROC
int_handler(void)
{
	int             istat;
	int             bstat;
	int             unit;
	int             (*f) (int, int, caddr_t);
	caddr_t         y;
	istat = inb(io_addr + INT_STATUS);
	while ((bstat = inb(io_addr + STATUS)) & BUSY);
	unit = istat & 0x0f;
	outb(io_addr + ATTENTION, unit | END_OF_INT);
	if (f = scsimap[unit].f) {
		y = scsimap[unit].ptr;
		scsimap[unit].f = 0;
		return (*f) (istat, bstat, y);
	}
	return 0;
}


send_command(const unsigned code, const long command)
{
	int             timeout;
	int             psw;

	timeout = 0;

	psw = disable();
	while (inb(io_addr + STATUS) & BUSY) {
		restore(psw);
		if (--timeout == 0) {
			outb(io_addr + CONTROL, 0);
			return -1;
		}
		psw = disable();
	}

	outb(io_addr + INTERFACE_CMD1, command);
	outb(io_addr + INTERFACE_CMD2, command >> 8);
	outb(io_addr + INTERFACE_CMD3, command >> 16);
	outb(io_addr + INTERFACE_CMD4, command >> 24);

	outb(io_addr + ATTENTION, code);

	restore(psw);
	return 0;
}

io_cmd(int func, daddr_t blockno, paddr_t dma_addr, int nblk)
{
	scb_ptr->enable = SCB_ENABLE_RETRY | SCB_STAT_ON_ERR;
	if ((scb_ptr->command = func) == READ_DATA) {
		scb_ptr->enable |= SCB_DATA_READ;
	}
	scb_ptr->laddr = blockno;
	scb_ptr->buffer = dma_addr;
	scb_ptr->block_cnt = nblk;
	scb_ptr->nbytes = _BLOCK_SIZE * nblk;
	scb_ptr->tsb = tsb_phys_addr;
	scb_ptr->chain = 0;
	scb_ptr->block_len = _BLOCK_SIZE;
}

cap_cmd()
{
	scb_ptr->command = READ_CAPACITY;
	scb_ptr->enable = SCB_BYPASS_BUFFER | SCB_ENABLE_RETRY |
		SCB_STAT_ON_ERR | SCB_DATA_READ;
	scb_ptr->laddr = 0;
	scb_ptr->buffer = cap_phys_addr;
	scb_ptr->nbytes = sizeof(struct _capacity);
	scb_ptr->tsb = tsb_phys_addr;
	scb_ptr->chain = 0;
	scb_ptr->block_cnt = 0;
	scb_ptr->block_len = 0;
}
