
#ifndef _ps2scsi_h_included
#define _ps2scsi_h_included


#define MAX_IO         36
#define CTL_ADDRESS    0x3540
#define CTL_INTERRUPT  14
#define STACK_SIZE    256    
#define MAX_NDRIVES    7
#define MAX_PARALLEL   1

#define PHYS_DRIVE_SPEC            /* Override default of 0            */

#define PS2_NDEVS     16
#define ADAPTER_ID    15
#define INTR_STATUS(x)  (((x)>>4) & 0x0f)

/* Interrupt codes */
#define ADAPTER_RESET   0x00
#define COMMAND_OK      0x01
#define COMMAND_RETRY   0x02
#define ADAPTER_FAILURE 0x07
#define IMMEDIATE_OK    0x0a
#define COMMAND_ERROR   0x0c
#define COMMAND_INVALID 0x0e
#define SOFTWARE_FAULT  0x0f

/*-- PS/2 SCSI Hard Disk I/O assignments --*/
#define INTERFACE_CMD1 0    /* Read/Write 8    */
#define INTERFACE_CMD2 1    /* Read/Write 8    */
#define INTERFACE_CMD3 2    /* Read/Write 8    */
#define INTERFACE_CMD4 3    /* Read/Write 8    */
#define ATTENTION      4    /* Read/Write 8    */
#define CONTROL        5    /* Read/Write 8    */
#define INT_STATUS     6    /* Read 8        */
#define STATUS         7    /* Read 8        */

/*-- Commands for Attention Register --*/
#define IMMEDIATE_CMD  0x10    /* Command in registers            */
#define START_SCB      0x30    /* Command in memory             */
#define START_LONG_SCB 0x40    /* Command in memory            */
#define END_OF_INT     0xe0    /* End of interrupt                */

/*-- Control register bits --*/
#define DMA_ENABLE     0x02    /* Enable DMA operations        */
#define INT_ENABLE     0x01    /* Enable INT's to the bus        */
#define HARDWARE_RESET 0x80    /* Reset the adaptor            */

/*-- Status bits found in status register --*/
#define BUSY          0x01
#define INT_REQUEST   0x02
#define CMD_REG_EMPTY 0x04
#define CMD_REG_FULL  0x08

/*-- Controller Command Codes --*/
#define READ_DATA     0x1c01    /* Read sector                    */
#define WRITE_DATA    0x1c02    /* Write sector                    */
#define READ_CAPACITY 0x1c09    /* Get size of device            */

/*-- IO port and bit for flashing the hard disk light --*/
#define SYS_PORT_A    0x0092
#define DISK_LIGHT    0x40

/* SCB enable word options */
#define SCB_CHAINED_CMD    0x00001
#define SCB_BYPASS_BUFFER  0x00200
#define SCB_SHORTDATA_OK   0x00400
#define SCB_SCATGATH_TAB   0x01000
#define SCB_ENABLE_RETRY   0x02000
#define SCB_STAT_ON_ERR    0x04000
#define SCB_DATA_READ      0x08000
/*
 * This structure must be double word aligned. We therefore allocate
 * a buffer with 4 extra bytes and point scb_ptr into it so it is 
 * double word aligned.
 */
struct _scb {
    ushort_t   command,
               enable;
    paddr_t    laddr,
               buffer;
    long       nbytes;
    paddr_t    tsb,
               chain;
    ushort_t   block_cnt,
               block_len;
};

/*
 * This structure must be double word aligned. We therefore allocate
 * a buffer with 4 extra bytes and point into it.
 */
struct _capacity {
    long    nblocks;
    long    blksize;
};
    

/* device handler table */
struct hd_func {
     int (*f)(int istat, int bstat, caddr_t ptr);
     caddr_t ptr;
};

#endif
